import os.path
import py.path
import pytest

from yuio.parse import *


def test_str():
    parser = Str()
    assert parser('Test') == 'Test'
    assert parser.parse('Test') == 'Test'
    assert parser.parse_config('Test') == 'Test'
    with pytest.raises(ValueError, match='expected a string'):
        parser.parse_config(10)


def test_str_lower():
    parser = StrLower()
    assert parser('Test') == 'test'
    assert parser.parse('Test') == 'test'
    assert parser.parse_config('Test') == 'test'
    with pytest.raises(ValueError, match='expected a string'):
        parser.parse_config(10)


def test_str_upper():
    parser = StrUpper()
    assert parser('Test') == 'TEST'
    assert parser.parse('Test') == 'TEST'
    assert parser.parse_config('Test') == 'TEST'
    with pytest.raises(ValueError, match='expected a string'):
        parser.parse_config(10)


def test_int():
    parser = Int()
    assert parser('1') == 1
    assert parser.parse('1') == 1
    assert parser.parse_config(1) == 1
    assert parser.parse_config(1.0) == 1
    with pytest.raises(ValueError, match='could not parse'):
        parser.parse('x')
    with pytest.raises(ValueError, match='expected an int'):
        parser.parse_config(1.5)
    with pytest.raises(ValueError, match='expected an int'):
        parser.parse_config('x')


def test_float():
    parser = Float()
    assert parser('1.5') == 1.5
    assert parser('-10') == -10.0
    assert parser('2e9') == 2e9
    assert parser.parse_config(1.0) == 1.0
    assert parser.parse_config(1.5) == 1.5
    with pytest.raises(ValueError, match='could not parse'):
        parser.parse('x')
    with pytest.raises(ValueError, match='expected a float'):
        parser.parse_config('x')


def test_bool():
    parser = Bool()
    assert parser('y') is True
    assert parser('yes') is True
    assert parser('yEs') is True
    assert parser('n') is False
    assert parser('no') is False
    assert parser('nO') is False
    with pytest.raises(ValueError):
        parser.parse('Meh')
    assert parser.parse_config(True) is True
    assert parser.parse_config(False) is False
    with pytest.raises(ValueError, match='expected a bool'):
        parser.parse_config('x')

    assert parser.describe() == 'yes|no'
    assert parser.describe_value(True) == 'yes'
    assert parser.describe_value(False) == 'no'


def test_enum():
    class Cuteness(enum.Enum):
        CATS = enum.auto()
        DOGS = enum.auto()
        BLAHAJ = ':3'

    parser = Enum(Cuteness)
    assert parser('CATS') is Cuteness.CATS
    assert parser.parse('CATS') is Cuteness.CATS
    assert parser.parse_config('CATS') is Cuteness.CATS
    assert parser('dogs') is Cuteness.DOGS
    assert parser('Blahaj') is Cuteness.BLAHAJ
    with pytest.raises(ValueError):
        parser('Unchi')
    with pytest.raises(ValueError, match='expected a string'):
        parser.parse_config(10)

    assert parser.describe() == 'CATS|DOGS|BLAHAJ'
    assert parser.describe_value(Cuteness.BLAHAJ) == 'BLAHAJ'


def test_int_enum():
    class Colors(enum.IntEnum):
        RED = 31
        GREEN = 32
        BLUE = 34

    parser = Enum(Colors)
    assert parser('RED') is Colors.RED
    assert parser.parse('RED') is Colors.RED
    assert parser.parse_config('RED') is Colors.RED
    assert parser('green') is Colors.GREEN
    assert parser('Blue') is Colors.BLUE
    with pytest.raises(ValueError):
        parser('Color of a beautiful sunset')
    with pytest.raises(ValueError, match='expected a string'):
        parser.parse_config(10)

    assert parser.describe() == 'RED|GREEN|BLUE'
    assert parser.describe_value(Colors.RED) == 'RED'


def test_path():
    parser = Path()
    assert str(parser('/a/s/d')) == '/a/s/d'
    assert str(parser('/a/s/d/..')) == '/a/s'
    assert str(parser('a/s/d')) == os.path.abspath('a/s/d')
    assert str(parser('./a/s/./d')) == os.path.abspath('a/s/d')
    assert str(parser('~/a')) == os.path.expanduser('~/a')

    parser = Path(extensions=['.cfg', '.txt'])
    assert str(parser('/a/s/d.cfg')) == '/a/s/d.cfg'
    assert str(parser('/a/s/d.txt')) == '/a/s/d.txt'
    with pytest.raises(ValueError, match='should have extension .cfg, .txt'):
        parser('file.sql')


def test_file(tmpdir: py.path.local):
    tmpdir.join('file.cfg').write('hi!')

    parser = File()
    assert str(parser(tmpdir.join('file.cfg').strpath)) \
           == tmpdir.join('file.cfg').strpath
    with pytest.raises(ValueError, match='doesn\'t exist'):
        parser(tmpdir.join('file.txt').strpath)
    with pytest.raises(ValueError, match='is not a file'):
        parser(tmpdir.strpath)

    parser = File(extensions=['.cfg', '.txt'])
    assert str(parser(tmpdir.join('file.cfg').strpath)) \
           == tmpdir.join('file.cfg').strpath
    with pytest.raises(ValueError, match='doesn\'t exist'):
        parser(tmpdir.join('file.txt').strpath)
    with pytest.raises(ValueError, match='should have extension .cfg, .txt'):
        parser(tmpdir.join('file.sql').strpath)


def test_dir(tmpdir: py.path.local):
    tmpdir.join('file.cfg').write('hi!')

    parser = Dir()
    assert str(parser('~')) == os.path.expanduser('~')
    assert str(parser(tmpdir.strpath)) == tmpdir.strpath
    with pytest.raises(ValueError, match='doesn\'t exist'):
        parser(tmpdir.join('subdir').strpath)
    with pytest.raises(ValueError, match='is not a directory'):
        parser(tmpdir.join('file.cfg').strpath)


def test_git_repo(tmpdir: py.path.local):
    tmpdir.join('file.cfg').write('hi!')

    parser = GitRepo()
    with pytest.raises(ValueError, match='is not a git repository'):
        parser(tmpdir.strpath)
    with pytest.raises(ValueError, match='doesn\'t exist'):
        parser(tmpdir.join('subdir').strpath)
    with pytest.raises(ValueError, match='is not a directory'):
        parser(tmpdir.join('file.cfg').strpath)

    tmpdir.join('.git').mkdir()

    assert str(parser(tmpdir.strpath)) == tmpdir.strpath


def test_bound():
    with pytest.raises(TypeError, match='lower and lower_inclusive'):
        Bound(Int(), lower=0, lower_inclusive=1)
    with pytest.raises(TypeError, match='upper and upper_inclusive'):
        Bound(Int(), upper=0, upper_inclusive=1)

    parser = Bound(Int())

    assert parser('0') == 0
    assert parser('-10') == -10
    assert parser('10') == 10

    parser.lower_bound(0)
    assert parser('10') == 10
    with pytest.raises(ValueError, match='should be greater than 0'):
        parser('-1')
    with pytest.raises(ValueError, match='should be greater than 0'):
        parser('0')

    parser.lower_bound_inclusive(0)
    assert parser('10') == 10
    assert parser('0') == 0
    with pytest.raises(ValueError, match='should be greater or equal to 0'):
        parser('-1')

    parser.upper_bound(10)
    assert parser('5') == 5
    with pytest.raises(ValueError, match='should be lesser than 10'):
        parser('10')
    with pytest.raises(ValueError, match='should be lesser than 10'):
        parser('11')

    parser.upper_bound_inclusive(10)
    assert parser('5') == 5
    assert parser('10') == 10
    with pytest.raises(ValueError, match='should be lesser or equal to 10'):
        parser('11')

    parser = Bound(Int(), lower_inclusive=0, upper_inclusive=5)
    assert parser('0') == 0
    assert parser('2') == 2
    assert parser('5') == 5
    with pytest.raises(ValueError, match='should be greater or equal to 0'):
        parser('-1')
    with pytest.raises(ValueError, match='should be lesser or equal to 5'):
        parser('6')

    parser = Bound(Int(), lower=0, upper=5)
    assert parser('2') == 2
    with pytest.raises(ValueError, match='should be greater than 0'):
        parser('0')
    with pytest.raises(ValueError, match='should be lesser than 5'):
        parser('5')


def test_one_of():
    parser = OneOf(Str(), ['qux', 'duo'])
    assert parser('qux') == 'qux'
    assert parser('duo') == 'duo'
    with pytest.raises(ValueError, match="qux, duo"):
        parser('foo')
    with pytest.raises(ValueError, match="qux, duo"):
        parser('Qux')
    with pytest.raises(ValueError, match="qux, duo"):
        parser('Duo')

    assert parser.describe() == 'qux|duo'

    parser = OneOf(StrLower(), ['qux', 'duo'])
    assert parser('Qux') == 'qux'
    assert parser('Duo') == 'duo'


def test_regex():
    parser = Regex(Str(), r'^a|b$')
    assert parser('a') == 'a'
    assert parser('b') == 'b'
    with pytest.raises(ValueError, match=r"should match regex '\^a\|b\$'"):
        parser('foo')



# Yuio project, MIT licence.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides several parsers that convert (usually user-provided)
strings to python objects, or check user input, and throw
:class:`ValueError` in case of any issue.

All parsers are descendants of the :class:`Parser` class.
On the surface, they are just callables that take a string and return a python
object. That is, you can use them in any place that expects such callable,
for example in flags in argparse, or in :func:`~yuio.io.ask` method
from :mod:`yuio.io`.

When parsing fails, we raise :class:`ParsingError`.


Base parser
-----------

.. autoclass:: Parser
   :members:

.. autoclass:: ParsingError


Value parsers
-------------

.. autoclass:: Str

.. autoclass:: StrLower

.. autoclass:: StrUpper

.. autoclass:: Int

.. autoclass:: Float

.. autoclass:: Bool

.. autoclass:: Enum


File system path parsers
------------------------

.. autoclass:: Path

.. autoclass:: NonExistentPath

.. autoclass:: ExistingPath

.. autoclass:: File

.. autoclass:: Dir

.. autoclass:: GitRepo


Validators
----------

.. autoclass:: Bound

   .. automethod:: lower_bound

   .. automethod:: lower_bound_inclusive

   .. automethod:: upper_bound

   .. automethod:: upper_bound_inclusive

.. autoclass:: OneOf

.. autoclass:: Regex

"""

import abc
import argparse
import enum
import pathlib
import re
import typing as _t


class _Comparable(_t.Protocol):
    @abc.abstractmethod
    def __lt__(self, other) -> bool: ...

    @abc.abstractmethod
    def __gt__(self, other) -> bool: ...

    @abc.abstractmethod
    def __le__(self, other) -> bool: ...

    @abc.abstractmethod
    def __ge__(self, other) -> bool: ...

    @abc.abstractmethod
    def __eq__(self, other) -> bool: ...


T = _t.TypeVar('T')
C = _t.TypeVar('C', bound=_Comparable)
E = _t.TypeVar('E', bound=enum.Enum)


class ParsingError(ValueError, argparse.ArgumentTypeError):
    """Raised when parsing or validation fails.

    This exception is derived from both :class:`ValueError`
    and :class:`argparse.ArgumentTypeError` to ensure that error messages
    are displayed nicely with argparse, and handled correctly in other places.

    """


class Parser(_t.Generic[T], abc.ABC):
    """Base class for parsers.

    """

    def __init__(self):
        # For some reason PyCharm incorrectly derives some types without
        # this `__init__`. You can check with the following code:
        #
        # ```
        # def interact(msg: str, p: Parser[T]) -> T:
        #     return p(msg)
        #
        # val = interact('10', Int())  # type for `val` is not derived
        # ```
        pass

    def __call__(self, value: str) -> T:
        """Parse and verify user input, raise :class:`ParsingError` on failure.

        """

        parsed = self.parse(value)
        self.validate(parsed)
        return parsed

    @abc.abstractmethod
    def parse(self, value: str) -> T:
        """Parse user input, raise :class:`ParsingError` on failure.

        Don't forget to call :meth:`Parser.validate` after parsing a value.

        """

    @abc.abstractmethod
    def parse_config(self, value: _t.Any) -> T:
        """Parse value from a config, raise :class:`ParsingError` on failure.

        This method accepts python values, i.e. when parsing a json config.

        Don't forget to call :meth:`Parser.validate` after parsing a value.

        """

    @abc.abstractmethod
    def validate(self, value: T):
        """Verify parsed value, raise :class:`ParsingError` on failure.

        """

    def describe(self) -> _t.Optional[str]:
        """Return a human-readable description of an expected input.

        """

        return None

    def describe_value(self, value: T) -> _t.Optional[str]:
        """Return a human-readable description of a given value.

        """

        return None

    @classmethod
    @_t.no_type_check
    def from_type_hint(cls, ty: _t.Any) -> 'Parser[_t.Any]':
        """Create parser from a type hint.

        """

        origin = _t.get_origin(ty)
        args = _t.get_args(ty)

        if origin is _t.Optional:
            return cls.from_type_hint(args[0])
        elif origin is _t.Union and len(args) == 2 and args[1] is type(None):
            return cls.from_type_hint(args[0])
        elif origin is _t.Union and len(args) == 2 and args[0] is type(None):
            return cls.from_type_hint(args[1])
        elif ty is str:
            return Str()
        elif ty is int:
            return Int()
        elif ty is float:
            return Float()
        elif ty is bool:
            return Bool()
        elif isinstance(ty, type) and issubclass(ty, enum.Enum):
            return Enum(ty)
        else:
            raise TypeError(f'unsupported type {ty}')


class Str(Parser[str]):
    """Parser for str values.

    Applies a `modifier` to the value, if one is given.

    """

    def __init__(self, modifier: _t.Optional[_t.Callable[[str], str]] = None):
        super().__init__()

        self._modifier = modifier

    def parse(self, value: str) -> str:
        if self._modifier is not None:
            return self._modifier(value)
        else:
            return value

    def parse_config(self, value: _t.Any) -> str:
        if not isinstance(value, str):
            raise ParsingError('expected a string')
        if self._modifier is not None:
            return self._modifier(value)
        else:
            return value

    def validate(self, value: str):
        pass


class StrLower(Str):
    """Parser for str values that converts them to lowercase.

    """

    def __init__(self):
        super().__init__(str.lower)


class StrUpper(Str):
    """Parser for str values that converts them to uppercase.

    """

    def __init__(self):
        super().__init__(str.upper)


class Int(Parser[int]):
    """Parser for int values.

    """

    def parse(self, value: str) -> int:
        try:
            return int(value)
        except ValueError:
            raise ParsingError(f'could not parse value {value!r} as an int')

    def parse_config(self, value: _t.Any) -> int:
        if isinstance(value, float):
            if value != int(value):
                raise ParsingError('expected an int, got a float instead')
            value = int(value)
        if not isinstance(value, int):
            raise ParsingError('expected an int')
        return value

    def validate(self, value: int):
        pass


class Float(Parser[float]):
    """Parser for float values.

    """

    def parse(self, value: str) -> float:
        try:
            return float(value)
        except ValueError:
            raise ParsingError(f'could not parse value {value!r} as a float')

    def parse_config(self, value: _t.Any) -> float:
        if not isinstance(value, (float, int)):
            raise ParsingError('expected a float')
        return value

    def validate(self, value: float):
        pass


class Bool(Parser[bool]):
    """Parser for bool values, such as `'yes'` or `'no'`.

    """

    def parse(self, value: str) -> bool:
        value = value.lower()

        if value in ('y', 'yes', 'true', '1'):
            return True
        elif value in ('n', 'no', 'false', '0'):
            return False
        else:
            raise ParsingError(f'could not parse value {value!r},'
                             f' enter either \'yes\' or \'no\'')

    def parse_config(self, value: _t.Any) -> bool:
        if not isinstance(value, bool):
            raise ParsingError('expected a bool')
        return value

    def validate(self, value: bool):
        pass

    def describe(self) -> _t.Optional[str]:
        return 'yes|no'

    def describe_value(self, value: bool) -> _t.Optional[str]:
        return 'yes' if value else 'no'


class Enum(Parser[E]):
    """Parser for enums, as defined in the standard :mod:`enum` module.

    """

    def __init__(self, enum_type: _t.Type[E]):
        super().__init__()

        self._enum_type: _t.Type[E] = enum_type

    def parse(self, value: str) -> E:
        try:
            return self._enum_type[value.upper()]
        except KeyError:
            enum_values = ', '.join(e.name for e in self._enum_type)
            raise ParsingError(
                f'could not parse value {value!r}'
                f' as {self._enum_type.__name__},'
                f' should be one of {enum_values}')

    def parse_config(self, value: _t.Any) -> E:
        if not isinstance(value, str):
            raise ParsingError('expected a string')
        return self.parse(value)

    def validate(self, value: E):
        pass

    def describe(self) -> _t.Optional[str]:
        desc = '|'.join(e.name for e in self._enum_type)
        return desc

    def describe_value(self, value: E) -> _t.Optional[str]:
        return value.name


class Path(Parser[pathlib.Path]):
    """Parse a file system path, return a :class:`pathlib.Path`.

    """

    def __init__(self, extensions: _t.Optional[_t.Collection[str]] = None):
        """Init the class.

        :param extensions: list of allowed file extensions.

        """

        super().__init__()

        self._extensions = extensions

    def parse(self, value: str) -> pathlib.Path:
        return pathlib.Path(value).expanduser().resolve()

    def parse_config(self, value: _t.Any) -> pathlib.Path:
        if not isinstance(value, str):
            raise ParsingError('expected a string')
        return self.parse(value)

    def validate(self, value: pathlib.Path):
        if self._extensions is not None:
            if not any(value.name.endswith(ext) for ext in self._extensions):
                exts = ', '.join(self._extensions)
                raise ParsingError(f'{value} should have extension {exts}')

    def describe(self) -> _t.Optional[str]:
        if self._extensions is not None:
            return '|'.join('*' + e for e in self._extensions)
        else:
            return None


class NonExistentPath(Path):
    """Parse a file system path and verify that it doesn't exist.

    """

    def validate(self, value: pathlib.Path):
        super().validate(value)

        if value.exists():
            raise ParsingError(f'{value} already exist')


class ExistingPath(Path):
    """Parse a file system path and verify that it exists.

    """

    def validate(self, value: pathlib.Path):
        super().validate(value)

        if not value.exists():
            raise ParsingError(f'{value} doesn\'t exist')


class File(ExistingPath):
    """Parse path to a file.

    """

    def validate(self, value: pathlib.Path):
        super().validate(value)

        if not value.is_file():
            raise ParsingError(f'{value} is not a file')


class Dir(ExistingPath):
    """Parse path to a directory.

    """

    def __init__(self):
        # Disallow passing `extensions`.
        super().__init__()

    def validate(self, value: pathlib.Path):
        super().validate(value)

        if not value.is_dir():
            raise ParsingError(f'{value} is not a directory')


class GitRepo(Dir):
    """Parse path to a git repository.

    This parser just checks that the given directory has
    a subdirectory named ``.git``.

    """

    def validate(self, value: pathlib.Path):
        super().validate(value)

        if not value.joinpath('.git').is_dir():
            raise ParsingError(f'{value} is not a git repository')

        return value


class Bound(Parser[C]):
    """Check that value is upper- or lower-bound by some constraints.

    :param inner:
        inner parser that will be used to actually parse a value before
        checking its bounds.
    :param lower:
        set lower bound for value, so we require that ``value > lower``.
        Can't be given if `lower_inclusive` is also given.
    :param lower_inclusive:
        set lower bound for value, so we require that ``value >= lower``.
        Can't be given if `lower` is also given.
    :param upper:
        set upper bound for value, so we require that ``value < upper``.
        Can't be given if `upper_inclusive` is also given.
    :param upper_inclusive:
        set upper bound for value, so we require that ``value <= upper``.
        Can't be given if `upper` is also given.

    """

    _Self = _t.TypeVar('_Self', bound='Bound')

    _lower: _t.Optional[C] = None
    """Lower bound."""

    _lower_inclusive: bool = True
    """True if lower bound is inclusive."""

    _upper: _t.Optional[C] = None
    """Upper bound."""

    _upper_inclusive: bool = True
    """True if upper bound is inclusive."""

    def __init__(
        self,
        inner: Parser[C],
        *,
        lower: _t.Optional[C] = None,
        lower_inclusive: _t.Optional[C] = None,
        upper: _t.Optional[C] = None,
        upper_inclusive: _t.Optional[C] = None
    ):
        super().__init__()

        self._inner: Parser[C] = inner

        if lower is not None and lower_inclusive is not None:
            raise TypeError(
                'lower and lower_inclusive cannot be given at the same time')
        elif lower is not None:
            self.lower_bound(lower)
        elif lower_inclusive is not None:
            self.lower_bound_inclusive(lower_inclusive)

        if upper is not None and upper_inclusive is not None:
            raise TypeError(
                'upper and upper_inclusive cannot be given at the same time')
        elif upper is not None:
            self.upper_bound(upper)
        elif upper_inclusive is not None:
            self.upper_bound_inclusive(upper_inclusive)

    def lower_bound(self: _Self, lower: C) -> _Self:
        """Set lower bound so we require that `value > lower`.

        """

        self._lower = lower
        self._lower_inclusive = False
        return self

    def lower_bound_inclusive(self: _Self, lower: C) -> _Self:
        """Set lower bound so we require that `value >= lower`.

        """

        self._lower = lower
        self._lower_inclusive = True
        return self

    def upper_bound(self: _Self, upper: C) -> _Self:
        """Set upper bound so we require that `value < upper`.

        """

        self._upper = upper
        self._upper_inclusive = False
        return self

    def upper_bound_inclusive(self: _Self, upper: C) -> _Self:
        """Set upper bound so we require that `value <= upper`.

        """

        self._upper = upper
        self._upper_inclusive = True
        return self

    def parse(self, value: str) -> C:
        return self._inner.parse(value)

    def parse_config(self, value: _t.Any) -> C:
        return self._inner.parse_config(value)

    def validate(self, value: C):
        self._inner.validate(value)

        if self._lower is not None:
            if self._lower_inclusive and value < self._lower:
                raise ParsingError(
                    f'value should be greater or equal to {self._lower},'
                    f' got {value} instead')
            elif not self._lower_inclusive and value <= self._lower:
                raise ParsingError(
                    f'value should be greater than {self._lower},'
                    f' got {value} instead')

        if self._upper is not None:
            if self._upper_inclusive and value > self._upper:
                raise ParsingError(
                    f'value should be lesser or equal to {self._upper},'
                    f' got {value} instead')
            elif not self._upper_inclusive and value >= self._upper:
                raise ParsingError(
                    f'value should be lesser than {self._upper},'
                    f' got {value} instead')


class OneOf(Parser[T]):
    """Check if the parsed value is one of the given set of values.

    """

    def __init__(self, inner: Parser[T], values: _t.Collection[T]):
        super().__init__()

        self._inner = inner
        self._allowed_values = values

    def parse(self, value: str) -> T:
        return self._inner.parse(value)

    def parse_config(self, value: _t.Any) -> T:
        return self._inner.parse_config(value)

    def validate(self, value: T):
        self._inner.validate(value)

        if value not in self._allowed_values:
            values = ', '.join(map(str, self._allowed_values))
            raise ParsingError(
                f'could not parse value {value!r},'
                f' should be one of {values}')

    def describe(self) -> _t.Optional[str]:
        desc = '|'.join(str(e) for e in self._allowed_values)
        if len(desc) < 80:
            return desc
        else:
            return super().describe()


class Regex(Parser[str]):
    """Check if the parsed value is one of the given set of values.

    """

    def __init__(self, inner: Parser[str], regex: _t.Union[str, re.Pattern]):
        super().__init__()

        if isinstance(regex, str):
            regex = re.compile(regex)

        self._inner = inner
        self._regex = regex

    def parse(self, value: str) -> str:
        return self._inner.parse(value)

    def parse_config(self, value: _t.Any) -> str:
        return self._inner.parse_config(value)

    def validate(self, value: str):
        self._inner.validate(value)

        if self._regex.match(value) is None:
            raise ParsingError(
                f'value should match regex \'{self._regex.pattern}\'')



import enum
import os
import typing

import pytest

import yuio.config


def test_default():
    class Config(yuio.config.Config):
        f1: str = '1'
        f2: str = '2'

    c = Config()
    assert c.f1 == '1'
    assert c.f2 == '2'

    c = Config(f2='b')
    assert c.f1 == '1'
    assert c.f2 == 'b'

    c = Config(f1='a', f2='b')
    assert c.f1 == 'a'
    assert c.f2 == 'b'


def test_default_field():
    class Config(yuio.config.Config):
        f1: str = yuio.config.field('1')
        f2: str = yuio.config.field('2')

    c = Config()
    assert c.f1 == '1'
    assert c.f2 == '2'

    c = Config(f2='b')
    assert c.f1 == '1'
    assert c.f2 == 'b'

    c = Config(f1='a', f2='b')
    assert c.f1 == 'a'
    assert c.f2 == 'b'


def test_missing():
    class Config(yuio.config.Config):
        f1: str
        f2: str

    c = Config()
    with pytest.raises(AttributeError, match='f1 is not configured'):
        _ = c.f1
    with pytest.raises(AttributeError, match='f2 is not configured'):
        _ = c.f2

    c = Config(f1='a')
    assert c.f1 == 'a'
    with pytest.raises(AttributeError, match='f2 is not configured'):
        _ = c.f2

    c = Config(f1='a', f2='b')
    assert c.f1 == 'a'
    assert c.f2 == 'b'


def test_missing_field():
    class Config(yuio.config.Config):
        f1: str = yuio.config.field()
        f2: str = yuio.config.field()

    c = Config()
    with pytest.raises(AttributeError, match='f1 is not configured'):
        _ = c.f1
    with pytest.raises(AttributeError, match='f2 is not configured'):
        _ = c.f2

    c = Config(f1='a')
    assert c.f1 == 'a'
    with pytest.raises(AttributeError, match='f2 is not configured'):
        _ = c.f2

    c = Config(f1='a', f2='b')
    assert c.f1 == 'a'
    assert c.f2 == 'b'


def test_init():
    class Config(yuio.config.Config):
        f1: str
        f2: str

    with pytest.raises(TypeError, match=r'unknown field\(s\): x'):
        Config(x=1)

    with pytest.raises(TypeError, match=r'unknown field\(s\): x, y'):
        Config(f1='a', x=1, y=2)


def test_repr():
    class Config(yuio.config.Config):
        f1: str = '1'
        f2: str

    assert repr(Config()) == "Config(\n  f1='1'\n)"
    assert repr(Config(f1='x')) == "Config(\n  f1='x'\n)"
    assert repr(Config(f2='y')) == "Config(\n  f1='1',\n  f2='y'\n)"


def test_inheritance():
    class Parent(yuio.config.Config):
        field_without_default: int
        field_with_default: int = 0
        field_without_default_that_gets_overridden_with_default: int
        field_without_default_that_gets_overridden_without_default: int
        field_with_default_that_gets_overridden_with_default: int = 0
        field_with_default_that_gets_overridden_without_default: int = 0

    class Child(Parent):
        field_without_default_that_gets_overridden_with_default: int = 0
        field_without_default_that_gets_overridden_without_default: int
        field_with_default_that_gets_overridden_with_default: int = 0
        field_with_default_that_gets_overridden_without_default: int

        new_field: str = 'x'

    with pytest.raises(TypeError, match=r'unknown field\(s\): new_field'):
        Parent(new_field='-_-')

    c = Child(
        field_without_default=1,
        field_with_default=2,
        field_without_default_that_gets_overridden_with_default=3,
        field_without_default_that_gets_overridden_without_default=4,
        field_with_default_that_gets_overridden_with_default=5,
        field_with_default_that_gets_overridden_without_default=6,
        new_field='7',
    )

    assert c.field_without_default == 1
    assert c.field_with_default == 2
    assert c.field_without_default_that_gets_overridden_with_default == 3
    assert c.field_without_default_that_gets_overridden_without_default == 4
    assert c.field_with_default_that_gets_overridden_with_default == 5
    assert c.field_with_default_that_gets_overridden_without_default == 6
    assert c.new_field == '7'

    c = Child()

    with pytest.raises(AttributeError, match='is not configured'):
        _ = c.field_without_default
    assert c.field_with_default == 0
    assert c.field_without_default_that_gets_overridden_with_default == 0
    with pytest.raises(AttributeError, match='is not configured'):
        _ = c.field_without_default_that_gets_overridden_without_default
    assert c.field_with_default_that_gets_overridden_with_default == 0
    with pytest.raises(AttributeError, match='is not configured'):
        _ = c.field_with_default_that_gets_overridden_without_default
    assert c.new_field == 'x'


def test_update():
    class Config(yuio.config.Config):
        f1: str = 'f1'
        f2: str

    c = Config(f1='f1.2')
    c.update(Config(f1='f1.3'))
    assert c.f1 == 'f1.3'

    c = Config(f1='f1.2')
    c.update(Config())
    assert c.f1 == 'f1.2'

    c = Config(f1='f1.2')
    c.update(Config(f2='f2'))
    assert c.f1 == 'f1.2'
    assert c.f2 == 'f2'

    c = Config(f1='f1.2', f2='f2')
    c.update(Config(f2='f2.2'))
    assert c.f1 == 'f1.2'
    assert c.f2 == 'f2.2'


def test_load_from_env():
    os.environ.clear()

    class Config(yuio.config.Config):
        f1: str
        f2: str = 'f2'

    c = Config.load_from_env()
    with pytest.raises(AttributeError, match='is not configured'):
        _ = c.f1
    assert c.f2 == 'f2'

    os.environ['F1'] = 'f1.2'
    c = Config.load_from_env()
    assert c.f1 == 'f1.2'
    assert c.f2 == 'f2'

    os.environ['F2'] = 'f2.2'
    c = Config.load_from_env()
    assert c.f1 == 'f1.2'
    assert c.f2 == 'f2.2'


def test_load_from_env_disabled():
    os.environ.clear()

    class Config(yuio.config.Config):
        f_disabled: str = yuio.config.field(
            'f_disabled', env=yuio.config.disabled())
        f_enabled: str = 'f_enabled'

    os.environ['F_DISABLED'] = 'f_disabled.2'
    os.environ['F_ENABLED'] = 'f_enabled.2'
    c = Config.load_from_env()
    assert c.f_disabled == 'f_disabled'
    assert c.f_enabled == 'f_enabled.2'


def test_load_from_env_configured():
    os.environ.clear()

    class Config(yuio.config.Config):
        f1: str = yuio.config.field('f1', env='FX')

    os.environ['F1'] = 'f1.2'
    c = Config.load_from_env()
    assert c.f1 == 'f1'

    os.environ['FX'] = 'f1.3'
    c = Config.load_from_env()
    assert c.f1 == 'f1.3'


def test_load_from_env_parsers():
    os.environ.clear()

    class Config(yuio.config.Config):
        b: bool
        s: str = yuio.config.field(
            parser=yuio.parse.OneOf(yuio.parse.Str(), ['x', 'y'])
        )

    os.environ['B'] = 'y'
    os.environ['S'] = 'x'
    c = Config.load_from_env()
    assert c.b is True
    assert c.s is 'x'

    os.environ['S'] = 'z'
    with pytest.raises(ValueError, match='one of x, y'):
        _ = Config.load_from_env()

    os.environ.clear()

    class E(enum.Enum):
        A = enum.auto()
        B = enum.auto()

    class Config(yuio.config.Config):
        b: bool
        s: str
        f: float
        i: int
        e: E

    os.environ['B'] = 'no'
    os.environ['S'] = 'str'
    os.environ['F'] = '1.5'
    os.environ['I'] = '-10'
    os.environ['E'] = 'B'

    c = Config.load_from_env()
    assert c.b is False
    assert c.s == 'str'
    assert c.f == 1.5
    assert c.i == -10
    assert c.e is E.B

    os.environ.clear()

    class Config(yuio.config.Config):
        b: typing.Optional[bool]
        s: typing.Optional[str]
        f: typing.Optional[float]
        i: typing.Optional[int]
        e: typing.Optional[E]

    os.environ['B'] = 'no'
    os.environ['S'] = 'str'
    os.environ['F'] = '1.5'
    os.environ['I'] = '-10'
    os.environ['E'] = 'B'

    c = Config.load_from_env()
    assert c.b is False
    assert c.s == 'str'
    assert c.f == 1.5
    assert c.i == -10
    assert c.e is E.B


def test_load_from_config():
    pass


def test_load_from_config_unknown_options():
    pass


def test_load_from_config_unknown_options_ignored():
    pass


def test_load_from_config_parsers():
    pass



import io

from yuio.io import *


def test_color_output():
    assert str(Color()) == '\033[0m'
    assert str(Color(31)) == '\033[0;31m'
    assert str(Color(32, bold=True, dim=True)) == '\033[0;32;1;2m'
    assert str(Color(bold=True)) == '\033[0;1m'


def test_color_combine():
    assert Color(1) | Color(2) == Color(2)
    assert Color(1) | Color() == Color(1)
    assert Color() | Color(1) == Color(1)
    assert Color(1) | Color(bold=True) == Color(1, bold=True)
    assert Color(1, bold=True) | Color(2) == Color(2, bold=True)


def test_color_tags_are_removed_in_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    info('<c:t1>t1</c> <c:t2,t3>t2,t3</c>')

    assert stream.getvalue() == f't1 t2,t3\n'


def test_no_color_tags_property():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    info('<c:t1>t1</c>', extra={'no_color_tags': True})

    assert stream.getvalue() == f'<c:t1>t1</c>\n'


def test_simple_logging_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    debug('debug message')
    info('info message')
    warning('warning message')
    error('error message')

    assert stream.getvalue() == \
           f'debug message\n' \
           f'info message\n' \
           f'warning message\n' \
           f'error message\n'


def test_task_logging_ok_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task message')
    task_progress('task message', progress=0)
    task_progress('task message', progress=0.5)
    task_progress('task message', progress=1)
    task_done('task message')

    assert stream.getvalue() == \
           f'task message... OK\n'


def test_task_logging_err_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task message')
    task_progress('task message', progress=0)
    task_progress('task message', progress=0.5)
    task_progress('task message', progress=1)
    task_error('task message')

    assert stream.getvalue() == \
           f'task message... ERROR\n'


def test_multiline_task_logging_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task message line 1\ntask message line 2')
    task_progress('task message line 1\ntask message line 2', progress=0)
    task_progress('task message line 1\ntask message line 2', progress=1)
    task_done('task message line 1\ntask message line 2')

    assert stream.getvalue() == \
           f'task message line 1...\n' \
           f'task message line 2\n' \
           f'task message line 1... OK\n' \
           f'task message line 2\n'


def test_task_logging_interrupt_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task message')
    info('info message')
    task_done('task message')

    assert stream.getvalue() == \
           f'task message...\n' \
           f'info message\n' \
           f'task message... OK\n'


def test_task_logging_interrupt_with_other_task_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task 1 message')
    task_begin('task 2 message')
    task_done('task 1 message')
    task_done('task 2 message')

    assert stream.getvalue() == \
           f'task 1 message...\n' \
           f'task 2 message...\n' \
           f'task 1 message... OK\n' \
           f'task 2 message... OK\n'

    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task 1 message')
    task_begin('task 2 message')
    task_done('task 2 message')
    task_done('task 1 message')

    assert stream.getvalue() == \
           f'task 1 message...\n' \
           f'task 2 message... OK\n' \
           f'task 1 message... OK\n'


def test_task_logging_interrupt_with_question_no_colors():
    stream = io.StringIO()
    setup(use_colors=False, level=DEBUG, stream=stream)

    task_begin('task 1 message')
    question('prompt')

    assert stream.getvalue() == \
           f'task 1 message...\n' \
           f'prompt'


def test_simple_logging():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    debug('debug message')
    info('info message')
    warning('warning message')
    error('error message')

    assert stream.getvalue() == \
           f'\033[0;39;49;2mdebug message\n\033[0m' \
           f'\033[0;39;49minfo message\n\033[0m' \
           f'\033[0;33;49mwarning message\n\033[0m' \
           f'\033[0;31;49merror message\n\033[0m'


def test_color_tags():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    info('<c:red>red <c:bold>bold <c:green>green</c></c></c>')

    assert (
        stream.getvalue() ==
           f'\033[0;39;49m'  # info log line color, color is normal
           f'\033[0;31;49mred '  # opening red tag, color is red
           f'\033[0;31;49;1mbold '  # opening bold tag, color is bold-red
           f'\033[0;32;49;1mgreen'  # opening green tag, color is bold-green
           f'\033[0;31;49;1m'  # closing green tag
           f'\033[0;31;49m'  # closing bold tag
           f'\033[0;39;49m\n'  # closing red tag
           f'\033[0m'  # end of message, resetting all styles
    )


def test_color_overrides():
    stream = io.StringIO()
    setup(
        use_colors=True, level=DEBUG, stream=stream, colors=dict(
            info=STYLE_BOLD,
            custom_tag=FORE_BLACK,
        )
    )

    info('<c:custom_tag>black</c>')

    assert (
        stream.getvalue() ==
        f'\033[0;39;49;1m'  # info log line color, color is bold
        f'\033[0;30;49;1mblack'  # opening custom tag, color is bold-black
        f'\033[0;39;49;1m\n'  # closing custom tag
        f'\033[0m'  # end of message, resetting all styles
    )


def test_task_logging_ok():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    task_begin('task message')
    task_progress('task message', progress=0)
    task_progress('task message', progress=0.5)
    task_progress('task message', progress=1)
    task_done('task message')

    assert stream.getvalue() == \
           f'\033[0;34;49mtask message...\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [                    ] 0%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [=========>          ] 50%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [====================] 100%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... OK\n\033[0m'


def test_task_logging_err():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    task_begin('task message')
    task_progress('task message', progress=0)
    task_progress('task message', progress=0.5)
    task_progress('task message', progress=1)
    task_error('task message')

    assert stream.getvalue() == \
           f'\033[0;34;49mtask message...\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [                    ] 0%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [=========>          ] 50%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [====================] 100%\033[0m' \
           f'\033[2K\r\033[0;31;49mtask message... ERROR\n\033[0m'


def test_multiline_task_logging():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    task_begin('task message line 1\ntask message line 2')
    task_progress('task message line 1\ntask message line 2', progress=0.5)
    task_done('task message line 1\ntask message line 2')

    assert stream.getvalue() == \
           f'\033[0;34;49mtask message line 1...\n' \
           f'task message line 2\n\033[0m' \
           f'\033[0;34;49mtask message line 1... OK\n' \
           f'task message line 2\n\033[0m'


def test_task_interrupt():
    stream = io.StringIO()
    setup(use_colors=True, level=DEBUG, stream=stream)

    task_begin('task message')
    task_progress('task message', progress=0.5)
    info('info message')
    task_error('task message')

    assert stream.getvalue() == \
           f'\033[0;34;49mtask message...\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message... [=========>          ] 50%\033[0m' \
           f'\033[2K\r\033[0;34;49mtask message...\n\033[0m' \
           f'\033[0;39;49minfo message\n\033[0m' \
           f'\033[0;31;49mtask message... ERROR\n\033[0m'


def test_edit():
    os.environ['NON_INTERACTIVE'] = '1'

    text = edit(
        '# Comment\n'
        'Text'
    )

    assert text == 'Text'



# -- Project information -----------------------------------------------------

project = 'Yuio'
copyright = '2022, Tamika Nomara'
author = 'Tamika Nomara'


# -- General configuration ---------------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
]

templates_path = ['_templates']

exclude_patterns = []

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}

autodoc_typehints_format = 'short'
autodoc_member_order = 'bysource'

# -- Options for HTML output -------------------------------------------------

html_theme = 'sphinx_rtd_theme'



#!/usr/bin/env python3

import pathlib
import re

import yuio

RELEASE_HEADER = """# Please enter the release changelog. Lines starting
# with '#' will be ignored, and an empty changelog aborts the release.

"""


class Config(yuio.config.Config):
    repo_path: pathlib.Path = yuio.config.field(
        default=pathlib.Path(__file__).parent.parent.resolve(),
        parser=yuio.parse.GitRepo(),
        help='path to the git repo with project'
    )


def find_latest_version(repo: yuio.git.Repo):
    text = repo.git('tag', '--list', '--sort=-version:refname')
    tags = text.decode().split('\n')

    for tag in tags:
        match = re.match(r'^v(\d+)\.(\d+)\.(\d+)$', tag)
        if match is not None:
            return tag, f'v{match.group(1)}.{int(match.group(2)) + 1}.0'

    return None, None


def main():
    yuio.io.setup()

    config = Config()
    config.update(Config.load_from_env())
    config.update(Config.load_from_args())

    yuio.io.debug('config = %r', config)

    if not config.repo_path.joinpath('pyproject.toml').is_file():
        yuio.io.error(
            'File <c:code>pyproject.toml</c> wasn\'t found in the repository.'
        )
        exit(1)

    if not config.repo_path.joinpath('CHANGELOG.md').is_file():
        yuio.io.error(
            'File <c:code>CHANGELOG.md</c> wasn\'t found in the repository.'
        )
        exit(1)

    repo = yuio.git.Repo(config.repo_path)

    if repo.status().has_changes:
        yuio.io.error(
            'Your repository has uncommitted changes. '
            'Either commit them or stash them.'
        )
        exit(1)

    latest_version, next_version = find_latest_version(repo)

    head = repo.show('HEAD')
    assert head is not None

    commit = yuio.io.ask(
        'Enter the commit at which you want to cut the release',
        default=head,
        parser=yuio.git.RefParser(repo),
    )

    next_version = yuio.io.ask(
        'Enter the release tag',
        default=next_version,
        parser=yuio.parse.Regex(yuio.parse.Str(), r'^v\d+\.\d+\.\d+$')
    )

    branch = f'release/{next_version}'

    branch = yuio.io.ask(
        'Enter the name of the release branch',
        default=branch,
    )

    yuio.io.info('')
    yuio.io.info('Release parameters:')
    yuio.io.info('')
    yuio.io.info('Release version ..... <c:code>%s</c>', next_version)
    yuio.io.info('Release branch ...... <c:code>%s</c>', branch)
    yuio.io.info('Commit .............. <c:code>%s</c> (<c:code>%s</c>)',
                 commit.orig_ref, commit.short_hash)
    yuio.io.info('')
    yuio.io.info('Release changelog:')
    yuio.io.info('')

    changelog = ''
    if latest_version is not None:
        for entry in repo.log(f'{latest_version}..{commit}', max_entries=None):
            yuio.io.info(
                '- %s (by <c:note>%s</c>)', entry.title, entry.author
            )
            changelog += f'- {entry.title} (by {entry.author})\n'
    yuio.io.info('')

    if not yuio.io.ask('Do you want to proceed?', parser=yuio.parse.Bool()):
        yuio.io.error('Aborting release.')
        exit(1)

    if yuio.io.ask(
        'Do you want to edit changelog?',
        parser=yuio.parse.Bool(),
        default=True,
    ):
        changelog = yuio.io.edit(RELEASE_HEADER + changelog)
        if not changelog:
            yuio.io.error('Got empty changelog, aborting release.')
            exit(1)
        else:
            yuio.io.info('Changelog edit successful.')

    with yuio.io.Task('Checking out <c:code>%s</c>', commit.short_hash):
        repo.git('checkout', commit.hash)

    with yuio.io.Task('Creating release branch'):
        repo.git('branch', branch)
        repo.git('checkout', branch)

    with yuio.io.Task('Modifying <c:code>CHANGELOG.md</c>'):
        file = config.repo_path.joinpath('CHANGELOG.md')
        text = file.read_text()
        text = re.sub(
            r'^# Changelog\n',
            f'# Changelog\n\n*{next_version}:*\n\n{changelog}',
            text,
            flags=re.MULTILINE
        )
        file.write_text(text)

    with yuio.io.Task('Modifying <c:code>pyproject.toml</c>'):
        file = config.repo_path.joinpath('pyproject.toml')
        text = file.read_text()
        text = re.sub(
            r'^version\s*=\s*(?P<q>["\'])\d+\.\d+\.\d+(?P=q)$',
            f'version = "{next_version[1:]}"',
            text,
            flags=re.MULTILINE
        )
        file.write_text(text)

    yuio.io.info('Ready to create a release commit and tag.')
    yuio.io.info('Feel free to look around and add changes before proceeding.')

    if not yuio.io.ask('Do you want to proceed?', parser=yuio.parse.Bool()):
        yuio.io.error('Aborting release.')
        exit(1)

    with yuio.io.Task('Committing <c:code>pyproject.toml</c>'):
        repo.git('add', '.')
        repo.git('commit', '-m', f'Release {next_version}')

    with yuio.io.Task('Creating tag <c:code>%s</c>', next_version):
        repo.git('tag', next_version)

    yuio.io.info('Release branch is ready to be published.')

    if not yuio.io.ask(
        'Do you want to push the release branch '
        'to <c:code>origin</c>?',
        parser=yuio.parse.Bool()
    ):
        yuio.io.info(
            'Nothing more to do. '
            'Push the release branch <c:code>%s</c> and tag <c:code>%s</c> '
            'to origin to start deployment.',
            branch,
            next_version
        )
        exit(0)

    with yuio.io.Task('Pushing branch and tags to origin'):
        repo.git('push', '-u', 'origin', branch, next_version)

    yuio.io.info('<c:success>Release branch is published.</c>')
    yuio.io.info('You can view release progress in the github CI,')
    yuio.io.info(
        'see <c:code>'
        'https://github.com/taminomara/yuio/actions?query=branch:%s</c>',
        next_version
    )


if __name__ == '__main__':
    main()



# Yuio project, MIT licence.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module implements user-friendly input and output on top of the python's
standard logging library. It adds several special logging levels
for printing tasks with progress bars and other useful things.

Setup
-----

To init the logging system, call the :func:`setup` function. Ideally,
this function should be called at the very beginning of the program execution,
before any code had a chance to log anything. Calling :func:`setup` function
multiple times completely overrides all previous settings.

By default, :func:`setup` determines logging level and whether ANSI color codes
should be used based on environment variables and the state of the output
stream:

- ``DEBUG``: print debug-level messages,
- ``QUIET``: only print warnings, errors, and input prompts,
- ``NO_COLORS``: disable colored output,
- ``FORCE_COLORS``: enable colored output.

You can override this logic by providing your own arguments
to the :func:`setup` function.

.. autofunction:: setup


Logging messages
----------------

Use logging functions from this module (or from the python's :mod:`logging`
module):

.. autofunction:: debug

.. autofunction:: info

.. autofunction:: warning

.. autofunction:: error

.. autofunction:: exception

.. autofunction:: critical


Coloring the output
-------------------

By default, all log messages are colored according to their level.

If you need inline colors, you can use special tags in your log messages::

    info('Using the <c:code>code</c> tag.')

You can combine multiple colors in the same tag::

    info('<c:bold,green>Success!</c>')

To disable tags processing, pass `no_color_tags` flag
to the logging's `extra` object::

    info(
        'this tag --> <c:color> is printed as-is',
        extra=dict(no_color_tags=True)
    )

List of all tags available by default:

- ``code``: for inline code,
- ``note``: for notes, such as default values in user prompts,
- ``success``, ``failure``: for indicating outcome of the program,
- ``question``, ``critical``, ``error``, ``warning``, ``task``,
  ``task_done``, ``task_error``, ``info``, ``debug``:
  used to color log messages,
- ``bold``, ``dim``: font styles,
- ``red``, ``green``, ``yellow``, ``blue``, ``magenta``, ``cyan``, ``normal``:
  font colors.

You can add more tags or change colors of the existing ones by supplying
the `colors` argument to the :func:`setup` function. This argument
is a mapping from a tag name to a :class:`Color` instance::

    setup(
        colors=dict(
            success=FORE_BLUE | STYLE_BOLD
        )
    )

.. autoclass:: Color

List of all pre-defined codes:

.. autodata:: STYLE_BOLD

.. autodata:: STYLE_DIM

.. autodata:: FORE_NORMAL

.. autodata:: FORE_BLACK

.. autodata:: FORE_RED

.. autodata:: FORE_GREEN

.. autodata:: FORE_YELLOW

.. autodata:: FORE_BLUE

.. autodata:: FORE_MAGENTA

.. autodata:: FORE_CYAN

.. autodata:: FORE_WHITE

.. autodata:: BACK_NORMAL

.. autodata:: BACK_BLACK

.. autodata:: BACK_RED

.. autodata:: BACK_GREEN

.. autodata:: BACK_YELLOW

.. autodata:: BACK_BLUE

.. autodata:: BACK_MAGENTA

.. autodata:: BACK_CYAN

.. autodata:: BACK_WHITE


Indicating progress
-------------------

You can use the :class:`Task` class to indicate status and progress
of some task:

.. autoclass:: Task
   :members:


Querying user input
-------------------

To query some input from a user, there's the :func:`ask` function:

.. autofunction:: ask

You can prompt user to edit something with the :func:`edit` function.
Note that this function doesn't add any header with an explanation
to the given text, so you will need to do it yourself:

.. autofunction:: edit

There are some helper functions and classes:

.. autofunction:: is_interactive

.. autofunction:: detect_editor

.. autoclass:: UserIoError
   :members:


"""

import getpass
import logging
import os
import re
import string
import subprocess
import sys
import tempfile
import threading
import typing as _t
from dataclasses import dataclass

import yuio.parse

QUESTION = 100
CRITICAL = 50
ERROR = 40
WARNING = 30
TASK_BEGIN = 24
TASK_PROGRESS = 23
TASK_DONE = 22
TASK_ERROR = 21
INFO = 20
DEBUG = 10
NOTSET = 0


@dataclass(frozen=True)
class Color:
    """ANSI color code.

    See the list of all available codes
    at `Wikipedia <https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters>`_.

    Example::

        # 31 is a color code for red
        BOLD_RED = Color(fore=31, bold=True)

    You can combine multiple colors::

        BOLD_RED = FORE_RED | STYLE_BOLD

    """

    # Note: other font styles besides bold and dim are not implemented
    # because their support in terminals is limited.

    fore: _t.Optional[int] = None
    back: _t.Optional[int] = None
    bold: bool = False
    dim: bool = False

    def __or__(self, other: 'Color'):
        return Color(
            other.fore if other.fore is not None else self.fore,
            other.back if other.back is not None else self.back,
            self.bold or other.bold,
            self.dim or other.dim,
        )

    def __str__(self):
        codes = ['0']
        if self.fore is not None:
            codes.append(str(self.fore))
        if self.back is not None:
            codes.append(str(self.back))
        if self.bold:
            codes.append('1')
        if self.dim:
            codes.append('2')
        return '\033[' + ';'.join(codes) + 'm'


#: Bold font style.
STYLE_BOLD = Color(bold=True)
#: Dim font style.
STYLE_DIM = Color(dim=True)

#: Normal foreground color.
FORE_NORMAL = Color(fore=39)
#: Black foreground color.
FORE_BLACK = Color(fore=30)
#: Red foreground color.
FORE_RED = Color(fore=31)
#: Green foreground color.
FORE_GREEN = Color(fore=32)
#: Yellow foreground color.
FORE_YELLOW = Color(fore=33)
#: Blue foreground color.
FORE_BLUE = Color(fore=34)
#: Magenta foreground color.
FORE_MAGENTA = Color(fore=35)
#: Cyan foreground color.
FORE_CYAN = Color(fore=36)
#: White foreground color.
FORE_WHITE = Color(fore=37)

#: Normal background color.
BACK_NORMAL = Color(back=49)
#: Black background color.
BACK_BLACK = Color(back=40)
#: Red background color.
BACK_RED = Color(back=41)
#: Green background color.
BACK_GREEN = Color(back=42)
#: Yellow background color.
BACK_YELLOW = Color(back=43)
#: Blue background color.
BACK_BLUE = Color(back=44)
#: Magenta background color.
BACK_MAGENTA = Color(back=45)
#: Cyan background color.
BACK_CYAN = Color(back=46)
#: White background color.
BACK_WHITE = Color(back=47)

DEFAULT_FORMATTER = logging.Formatter(
    '%(message)s'
)

DEFAULT_COLORS = {
    'code': FORE_MAGENTA,
    'note': FORE_GREEN,
    'success': FORE_GREEN | STYLE_BOLD,
    'failure': FORE_RED | STYLE_BOLD,

    'question': FORE_BLUE,
    'critical': FORE_WHITE | BACK_RED,
    'error': FORE_RED,
    'warning': FORE_YELLOW,
    'task': FORE_BLUE,
    'task_done': FORE_BLUE,
    'task_error': FORE_RED,
    'info': FORE_NORMAL,
    'debug': STYLE_DIM,

    'bold': STYLE_BOLD,
    'dim': STYLE_DIM,
    'red': FORE_RED,
    'green': FORE_GREEN,
    'yellow': FORE_YELLOW,
    'blue': FORE_BLUE,
    'magenta': FORE_MAGENTA,
    'cyan': FORE_CYAN,
    'normal': FORE_NORMAL,
}

T = _t.TypeVar('T')

_HANDLER_LOCK = threading.Lock()
_HANDLER: _t.Optional['_Handler'] = None


def setup(
    level: _t.Optional[int] = None,
    stream: _t.Any = sys.stderr,
    use_colors: _t.Optional[bool] = None,
    formatter: logging.Formatter = DEFAULT_FORMATTER,
    colors: _t.Optional[_t.Dict[str, Color]] = None,
):
    """Initial setup of the logging facilities.

    :param level:
        log output level. If not given, will check ``DEBUG`` and ``QUIET``
        environment variables to determine an appropriate logging level.
    :param stream:
        a stream where to output messages. Uses `stderr` by default.
    :param use_colors:
        use ANSI escape sequences to color the output. If not given, will
        check ``NO_COLORS`` and ``FORCE_COLORS`` environment variables,
        and also if the given `stream` is a tty one.
    :param formatter:
        formatter for log messages.
    :param colors:
        mapping from tag name or logging level name to a :class:`Color`.
        Logging level names and tag names are all lowercase.

    """

    global _HANDLER

    with _HANDLER_LOCK:
        if _HANDLER is None:
            _HANDLER = _Handler()

            logger = logging.getLogger()
            logger.addHandler(_HANDLER)
            logger.setLevel(NOTSET)

            logging.addLevelName(QUESTION, "QUESTION")
            logging.addLevelName(TASK_BEGIN, "TASK")
            logging.addLevelName(TASK_PROGRESS, "TASK")
            logging.addLevelName(TASK_DONE, "TASK_DONE")
            logging.addLevelName(TASK_ERROR, "TASK_ERROR")

        if level is None:
            if 'DEBUG' in os.environ:
                level = DEBUG
            elif 'QUIET' in os.environ:
                level = WARNING
            else:
                level = INFO

        _HANDLER.setLevel(level)
        _HANDLER.setFormatter(formatter)
        _HANDLER.setStream(stream)

        full_colors = DEFAULT_COLORS.copy()
        if colors is not None:
            full_colors.update(colors)

        _HANDLER.setColors(full_colors)

        if use_colors is None:
            if 'NO_COLORS' in os.environ:
                use_colors = False
            elif 'FORCE_COLORS' in os.environ:
                use_colors = True
            else:
                use_colors = stream.isatty()

        _HANDLER.setUseColors(use_colors)


def debug(msg: str, *args, **kwargs):
    """Log a debug message.

    """

    logging.debug(msg, *args, **kwargs)


def info(msg: str, *args, **kwargs):
    """Log an info message.

    """

    logging.info(msg, *args, **kwargs)


def warning(msg: str, *args, **kwargs):
    """Log a warning message.

    """

    logging.warning(msg, *args, **kwargs)


def error(msg: str, *args, **kwargs):
    """Log an error message.

    """

    logging.error(msg, *args, **kwargs)


def exception(msg: str, *args, exc_info=True, **kwargs):
    """Log an error message and capture the current exception.

    Call this function in the `except` clause of a `try` block
    or in an `__exit__` function of a context manager to attach
    current exception details to the log message.

    """

    logging.exception(msg, *args, exc_info=exc_info, **kwargs)


def critical(msg: str, *args, **kwargs):
    """Log a critical error message.

    """

    logging.critical(msg, *args, **kwargs)


def task_begin(msg: str, *args):
    """Indicate beginning of some task.

    Note: prefer using the :class:`Task` context manager
    instead of this function.

    """

    logging.log(TASK_BEGIN, msg, *args)


def task_progress(msg: str, *args, progress: float):
    """Indicate progress of some task.

    If this function is called after :func:`task_begin`, the task message will
    be updated with a progress bar. If there were other log messages
    between :meth:`task_begin` and call to this function, nothing will happen.

    Note: prefer using the :class:`Task` context manager
    instead of this function.

    """

    extra = {'progress': progress}
    logging.log(TASK_PROGRESS, msg, *args, extra=extra)


def task_done(msg: str, *args):
    """Indicate that some task is finished successfully.

    Note: prefer using the :class:`Task` context manager
    instead of this function.

    """

    logging.log(TASK_DONE, msg, *args)


def task_error(msg: str, *args):
    """Indicate that some task is finished with an error.

    Note: prefer using the :class:`Task` context manager
    instead of this function.

    """

    logging.log(TASK_ERROR, msg, *args)


def task_exception(msg: str, *args):
    """Indicate that some task is finished, capture current exception.

    Note: prefer using the :class:`Task` context manager
    instead of this function.

    """

    logging.log(TASK_ERROR, msg, *args, exc_info=True)


def question(msg: str, *args):
    """Log a message with input prompts and other user communications.

    These messages don't end with newline. They also have high priority,
    so they will not be filtered by log level settings.

    """

    logging.log(QUESTION, msg, *args)


class Task:
    """Context manager that automatically reports tasks and their progress.

    This context manager accepts a name of the task being done,
    along with all the usual logging parameters.

    When context is entered, it emits a message that the task is in progress.

    When context is exited, it emits a message that the task has finished.

    If an exception is thrown out of context, it emits a message that the
    task has failed.

    Example::

        with Task('Fetching data') as task:
            for i, url in enumerate(urls):
                url.fetch()
                task.progress(float(i) / len(urls))

    This will output the following:

    .. code-block:: text

       Fetching data... [=====>              ] 30%

    Note that you don't have to use it in a context::

        task = Task('Fetching data')
        task.begin()

        for i, url in enumerate(urls):
            url.fetch()
            task.progress(float(i) / len(urls))

        task.done()

    In this case, however, it's up to you to catch and report errors.

    """

    def __init__(self, msg: str, *args):
        if args:
            msg = msg % args
        self._msg = msg

        self._started = False
        self._finished = False

    def begin(self):
        """Emit a message about task being in progress.

        """

        if not self._started:
            task_begin(self._msg)
            self._started = True

    def progress(self, progress: float):
        """Indicate progress of ths task, measured from ``0`` to ``1``.

        """

        if self._started and not self._finished:
            task_progress(self._msg, progress=progress)

    def done(self):
        """Emit a message that the task is finished successfully.

        """

        if self._started and not self._finished:
            task_done(self._msg)
            self._finished = True

    def error(self):
        """Emit a message that the task is finished with an error.

        """

        if self._started and not self._finished:
            task_error(self._msg)
            self._finished = True

    def exception(self):
        """Emit a message that the task is finished, capture current exception.

        """

        if self._started and not self._finished:
            task_exception(self._msg)
            self._finished = True

    def __enter__(self):
        self.begin()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is None:
            self.done()
        else:
            self.error()


class _Handler(logging.Handler):
    def __init__(self):
        super().__init__()

        self._use_colors: bool = sys.stderr.isatty()
        self._stream: _t.Any = sys.stderr
        self._current_task: _t.Optional[logging.LogRecord] = None
        self._colors = DEFAULT_COLORS

        self.setFormatter(logging.Formatter('-> %(message)s'))

    def setStream(self, stream):
        self._stream = stream

    def getStream(self):
        return self._stream

    def setUseColors(self, use_colors: bool):
        self._use_colors = use_colors

    def setColors(self, colors: _t.Dict[str, Color]):
        self._colors = colors

    def flush(self):
        self.acquire()
        try:
            self._stream.flush()
        finally:
            self.release()

    def emit(self, record: logging.LogRecord):
        record.message = record.getMessage()

        message = self._colorize(self.format(record), record)
        is_current_task = (
            record.levelno in (TASK_PROGRESS, TASK_DONE, TASK_ERROR)
            and self._current_task is not None
            and record.message == self._current_task.message
        )

        if record.levelno == TASK_DONE:
            status = 'OK'
            progress = None
        elif record.levelno == TASK_ERROR:
            status = 'ERROR'
            progress = None
        elif record.levelno in (TASK_BEGIN, TASK_PROGRESS):
            status = ''
            progress = getattr(record, 'progress', None)
        else:
            status = None
            progress = None

        if is_current_task:
            # When we're updating current task, the console looks something
            # like this:
            #
            # ```
            # -> Some previous message.
            # -> Some previous message.
            # -> Task description...|
            # ```
            #                       ^ cursor is at the end of the current line
            #
            # If we don't support colors, the only thing we can do is to print
            # task's status after the ellipsis and be done with it.
            #
            # If we support colors, we clear the current line, and redraw
            # the task message completely, including its current status
            # and progress.
            #
            # If the new message is multiline, though, we won't be able
            # to clear it and redraw next time. So we print it with a newline
            # at the end and clear the current task.

            if not self._use_colors and record.levelno == TASK_PROGRESS:
                # There is nothing we can do to indicate progress w/o colors.
                pass
            elif not self._use_colors:
                # Write status after the ellipsis and clear the task.
                self._stream.write(f' {status}\n')
                self._current_task = None
            else:
                # Redraw the current task.
                self._clear_line()
                is_multiline = self._print_task(message, status, progress)

                if record.levelno != TASK_PROGRESS or is_multiline:
                    # If new task record is multiline, we won't be able
                    # to redraw it next time, so we need to clear the task.
                    # If the task is finished, we also need to clear it.
                    self._stream.write('\n')
                    self._current_task = None
        elif record.levelno == TASK_PROGRESS:
            # We can't update status of a non-current task, so just ignore it.
            return
        else:
            if self._current_task is not None:
                # We need to clear current task before printing
                # new log messages.

                if self._use_colors:
                    # If we are using colors, we want to clear progress bar
                    # and leave just a task description with an ellipsis.
                    current_task_message = self._colorize(
                        self.format(self._current_task),
                        self._current_task)
                    self._clear_line()
                    self._print_task(current_task_message, '', None)

                self._stream.write('\n')
                self._current_task = None
                self._reset_color()

            is_multiline = self._print_task(message, status, progress)

            if record.levelno == TASK_BEGIN and not is_multiline:
                # If we've started a new task, mark is as current.
                # If this task's message is multiline, though,
                # don't mark it as current, and treat it as a regular
                # log message, though, because we can't redraw multiline tasks.
                self._current_task = record
            elif record.levelno != QUESTION:
                self._stream.write('\n')

        self._reset_color()

        self.flush()

    def _print_task(
        self,
        message: str,
        status: _t.Optional[str],
        progress: _t.Optional[float],
    ) -> bool:
        lines = message.split('\n')
        is_multiline = len(lines) > 1

        self._stream.write(lines[0])

        if status is not None:
            # Status is set for any task. If status is an empty line,
            # the task is in progress, so we only need an ellipsis.
            self._stream.write('...')
            if status:
                self._stream.write(' ' + status)

        if progress is not None and not is_multiline:
            # Never print progressbar for multiline messages because
            # we won't be able to update this progress bar in the future.
            self._stream.write(' ')
            self._stream.write(self._make_progress_bar(progress))

        if is_multiline:
            self._stream.write('\n')
            self._stream.write('\n'.join(lines[1:]))

        return is_multiline

    def _clear_line(self):
        if self._use_colors:
            self._stream.write(f'\033[2K\r')

    def _reset_color(self):
        if self._use_colors:
            self._stream.write(str(Color()))

    _TAG_RE = re.compile(r'<c:(?P<name>[a-z0-9, _-]+)>|</c>')

    def _colorize(self, msg: str, record: logging.LogRecord):
        if getattr(record, 'no_color_tags', False):
            return msg

        if not self._use_colors:
            return self._TAG_RE.sub('', msg)

        default_color = self._colors.get(record.levelname.lower(), Color())
        default_color = FORE_NORMAL | BACK_NORMAL | default_color

        out = ''
        last_pos = 0
        stack = [default_color]

        out += str(default_color)

        for tag in self._TAG_RE.finditer(msg):
            out += msg[last_pos:tag.span()[0]]
            last_pos = tag.span()[1]

            name = tag.group('name')
            if name:
                color = stack[-1]
                for sub_name in name.split(','):
                    sub_name = sub_name.strip()
                    color = color | self._colors.get(sub_name, Color())
                out += str(color)
                stack.append(color)
            elif len(stack) > 1:
                stack.pop()
                out += str(stack[-1])

        out += msg[last_pos:]

        return out

    @staticmethod
    def _make_progress_bar(progress: float):
        if progress < 0:
            progress = 0
        elif progress > 1:
            progress = 1

        done = int(20 * progress)
        infl = 0
        left = 20 - done

        if 0 < done < 20:
            infl = 1
            done -= 1

        return '[' \
               + '=' * done \
               + '>' * infl \
               + ' ' * left \
               + ']' \
               + f' {progress:0.0%}'


class UserIoError(IOError):
    """Raised when interaction with user fails.

    """


def is_interactive() -> bool:
    """Check if we're running in an interactive environment.

    """

    with _HANDLER_LOCK:
        if _HANDLER is None:
            return False

        return _HANDLER.getStream().isatty() \
            and os.environ.get('TERM', None) != 'dumb' \
            and 'NON_INTERACTIVE' not in os.environ \
            and 'CI' not in os.environ


@_t.overload
def ask(
    msg: str,
    *args,
    default: _t.Optional[str] = None,
    input_description: _t.Optional[str] = None,
    default_description: _t.Optional[str] = None,
    hidden: bool = False,
) -> str: ...


@_t.overload
def ask(
    msg: str,
    *args,
    parser: yuio.parse.Parser[T],
    default: _t.Optional[T] = None,
    input_description: _t.Optional[str] = None,
    default_description: _t.Optional[str] = None,
    hidden: bool = False,
) -> T: ...


def ask(
    msg: str,
    *args,
    parser: _t.Optional[yuio.parse.Parser[T]] = None,
    default: _t.Optional[T] = None,
    input_description: _t.Optional[str] = None,
    default_description: _t.Optional[str] = None,
    hidden: bool = False,
):
    """Ask user to provide an input, then parse it and return a value.

    If launched in a non-interactive environment, returns the default
    if one is present, or raises a :class:`UserIoError`.

    Example::

        answer = ask(
            'Do you want a choco bar?',
            parser=yuio.parse.Bool(),
            default=True,
        )

    :param msg:
        prompt that will be sent to the user.
    :param args:
        arguments for prompt formatting.
    :param parser:
        how to parse and verify the input. See :mod:`yuio.parse` for more
        info.
    :param default:
        if given, this value will be returned if no input is provided.
    :param input_description:
        a string that describes expected input, like ``'yes/no'`` for boolean
        inputs. By default, inferred from the given parser.
    :param default_description:
        a string that describes the `default` value. By default,
        inferred from the given parser and `repr` of the default value.
    :param hidden:
        if enabled, treats input as password, and uses secure input methods.
        This option also hides errors from the parser, because they may contain
        user input.

    """

    if not is_interactive():
        if default is not None:
            return default
        else:
            raise UserIoError(
                'can\'t interact with user in non-interactive environment'
            )

    if parser is None:
        parser = _t.cast(yuio.parse.Parser[T], str)

    desc = ''

    if input_description is None and hasattr(parser, 'describe'):
        input_description = getattr(parser, 'describe')()
    if input_description:
        desc += f' ({input_description})'

    if default is not None:
        if default_description is None and hasattr(parser, 'describe_value'):
            default_description = getattr(parser, 'describe_value')(default)
        if default_description is None:
            default_description = str(default)
        if default_description:
            desc += f' [<c:note>{default_description}</c>]'

    if args:
        msg = msg % args

    msg += desc

    if desc or not msg.endswith(tuple(string.punctuation)):
        msg += ': '
    else:
        msg += ' '

    while True:
        question(msg)
        try:
            if hidden:
                answer = getpass.getpass(prompt='')
            else:
                answer = input()
        except EOFError:
            raise UserIoError('unexpected end of input')
        if not answer and default is not None:
            return default
        elif not answer:
            error('Input is required.')
        else:
            try:
                return parser(answer)
            except Exception as e:
                if hidden:
                    error('Error: invalid value.')
                else:
                    error(f'Error: {e}.')


def detect_editor() -> _t.Optional[str]:
    """Detect an editor executable.

    This function checks the ``EDITOR`` environment variable.
    If it's not found, it checks whether ``nano`` or ``vi``
    are available. Otherwise, it returns `None`.

    """

    if 'EDITOR' in os.environ:
        return os.environ['EDITOR']
    elif (
        subprocess.run(
            ['which', 'nano'], stdout=subprocess.DEVNULL
        ).returncode == 0
    ):
        return 'nano'
    elif (
        subprocess.run(
            ['which', 'vi'], stdout=subprocess.DEVNULL
        ).returncode == 0
    ):
        return 'vi'
    else:
        return None


def edit(
    text: str,
    comment_marker: _t.Optional[str] = '#',
    editor: _t.Optional[str] = None,
) -> str:
    """Ask user to edit some text.

    This function creates a temporary file with the given text
    and opens it in an editor. After editing is done, it strips away
    all lines that start with `comment_marker`, if one is given.

    If editor is not available or returns a non-zero exit code,
    a :class:`UserIoError` is raised.

    If launched in a non-interactive environment, returns the text
    unedited (comments are still removed, though).

    """

    if is_interactive():
        if editor is None:
            editor = detect_editor()

        if editor is None:
            raise UserIoError(
                'can\'t detect an editor, ensure that the $EDITOR environment '
                'variable contains a correct path to an editor executable'
            )

        filepath = tempfile.mktemp()
        with open(filepath, 'w') as file:
            file.write(text)

        try:
            try:
                res = subprocess.run([editor, filepath])
            except FileNotFoundError:
                raise UserIoError(
                    'can\'t use this editor, ensure that the $EDITOR '
                    'environment variable contains '
                    'a correct path to an editor executable'
                )

            if res.returncode != 0:
                raise UserIoError(
                    'editing failed'
                )

            with open(filepath, 'r') as file:
                text = file.read()
        finally:
            os.remove(filepath)

    if comment_marker is not None:
        text = re.sub(
            r'^\s*' + re.escape(comment_marker) + r'.*?\n',
            '',
            text,
            flags=re.MULTILINE
        )

    return text.strip()



# Yuio project, MIT licence.
#
# https://github.com/taminomara/yuio/
#
# You're free to copy this file to your project and edit it for your needs,
# just keep this copyright line please :3

"""
This module provides base a class for configs that can be loaded from
files, environment variables or command line arguments.


Creating and loading configs
----------------------------

Derive your config from the :class:`Config` base class. Inside of its body,
define config fields using type annotations. This is similar to how you
create dataclasses::

    class AppConfig(Config):
        app_version: str
        deployment_strategy: str = 'ROLLING'

:class:`Config` defines `__init__` that takes keyword arguments for defined
fields, as well as some constructors that load config from environment
variables, files, and so on.

Upon creation, all fields that aren't explicitly initialized and don't have
defaults are considered missing. Accessing them
will raise :class:`AttributeError`.

Here's an example of loading our config from command line arguments
and environment variables::

    # Load config from env.
    config = AppConfig.load_from_env()

    # Create and setup parser.
    parser = argparse.ArgumentParser()
    AppConfig.setup_parser(parser)

    # Parse arguments, convert them to config,
    # and merge with config loaded from env.
    config.update(AppConfig.load_from_args(parser.parse_args()))

.. autoclass:: Config
   :members:


Advanced field configuration
----------------------------

By default, :class:`Config` infers names for env variables and flags,
appropriate parsers, and other things from field's name and type hint.
If you need to override them, theres the :func:`field` function::

    class AppConfig(Config):
        app_version: str = field(
            default='HEAD',
            help='git tag or commit hash at which a new version is built.',
        )
        deployment_strategy: str = field(
            default='ROLLING',
            help='strategy that will be used to deploy new pods.',
            parser=yuio.parse.OneOf(
                yuio.parse.StrUpper(),
                ['ROLLING', 'READONLY', 'DOWNTIME']
            )

.. autofunction:: field

.. autofunction:: disabled

"""

import argparse
import json
import os
import typing as _t
from dataclasses import dataclass

import yuio.parse


T = _t.TypeVar('T')


class _Placeholder:
    def __init__(self, what: str):
        self._what = what

    def __bool__(self):
        return False

    def __repr__(self):
        return f'<{self._what}>'


_MISSING: _t.Any = _Placeholder('value is missing')
_DISABLED: _t.Any = _Placeholder('option is disabled')


def disabled() -> _t.Any:
    """Placeholder indicating that some field's functionality is disabled.

    Example::

        class AppConfig(Config):
            app_version: str = field(env=disabled())

    """

    return _DISABLED


@dataclass(frozen=True)
class _FieldSettings:
    default: _t.Any = _MISSING
    parser: _t.Optional[yuio.parse.Parser] = None
    help: _t.Optional[str] = None
    env: _t.Optional[str] = None
    flag: _t.Optional[_t.Union[str, _t.List[str]]] = None

    def _update_defaults(self, qualname, name, ty) -> '_Field':
        default = self.default

        parser = self.parser
        if parser is None:
            try:
                parser = yuio.parse.Parser.from_type_hint(ty)
            except TypeError:
                raise TypeError(
                    f'can\'t derive parser for {qualname}.{name}')

        help = self.help
        if help is None:
            help = 'not documented.'

        env = self.env
        if env is None:
            env = name.upper()

        flag = self.flag
        if flag is None:
            flag = '--' + name.replace('_', '-')
        if not isinstance(flag, list):
            flag = [flag]

        dest = qualname.replace('.', '__') + '__' + name

        return _Field(
            default,
            parser,
            help,
            env,
            flag,
            dest
        )


@dataclass(frozen=True)
class _Field:
    default: _t.Any
    parser: yuio.parse.Parser
    help: str
    env: str
    flag: _t.List[str]
    dest: str


@_t.overload
def field(
    *,
    help: _t.Optional[str] = None,
    env: _t.Optional[str] = None,
    flag: _t.Optional[_t.Union[str, _t.List[str]]] = None,
) -> _t.Any: pass


@_t.overload
def field(
    default: T = _MISSING,
    *,
    parser: yuio.parse.Parser[T],
    help: _t.Optional[str] = None,
    env: _t.Optional[str] = None,
    flag: _t.Optional[_t.Union[str, _t.List[str]]] = None,
) -> T: pass


def field(
    default: T = _MISSING,
    *,
    parser: _t.Optional[yuio.parse.Parser[T]] = None,
    help: _t.Optional[str] = None,
    env: _t.Optional[str] = None,
    flag: _t.Optional[_t.Union[str, _t.List[str]]] = None,
) -> T:
    """Field descriptor, used for additional configuration of fields.

    This is similar to what :func:`dataclasses.field` does.

    :param default:
        default value for config field, used if field is missing from config.
    :param parser:
        parser that will be used to parse env vars, configs and CLI arguments.
        By default, it's inferred from type hint.
    :param help:
        help message that will be used in CLI argument description.
    :param env:
        name of environment variable that will be used for this field.
        By default, it's inferred from field name.
        Set to :func:`disabled` to disable parsing from environment variable.
    :param flag:
        name of a CLI flag (or a list of names) that will be used
        for this field. By default, it's inferred from field name.
        Set to :func:`disabled` to disable parsing from command line arguments.

    """

    return _t.cast(_t.Any, _FieldSettings(
        default=default,
        parser=parser,
        help=help,
        env=env,
        flag=flag
    ))


class _ConfigMeta(type):
    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)

        fields = {}

        for base in reversed(cls.__mro__):
            if hasattr(base, '_fields'):
                fields.update(getattr(base, '_fields'))

        for name, ty in cls.__annotations__.items():
            if name.startswith('_'):
                continue

            value = namespace.get(name, _MISSING)
            if isinstance(value, _FieldSettings):
                field = value
            else:
                field = _FieldSettings(default=value)
            setattr(cls, name, field.default)

            fields[name] = field._update_defaults(cls.__qualname__, name, ty)

        cls._fields = fields


class Config(metaclass=_ConfigMeta):
    """Base class for configs.

    """

    _Self = _t.TypeVar('_Self', bound='Config')

    # Value is generated by metaclass.
    _fields: _t.ClassVar[_t.Dict[str, _Field]]

    def __init__(self, **kwargs):
        for name in self._fields:
            if name in kwargs:
                setattr(self, name, kwargs.pop(name))

        if kwargs:
            unknown_fields = ', '.join(kwargs)
            raise TypeError(f'unknown field(s): {unknown_fields}')

    def update(self: _Self, other: _Self):
        """Update fields in this config with fields from another config.

        This function is similar to :meth:`dict.update`.

        """

        for name in self._fields:
            if name in other.__dict__:
                setattr(self, name, getattr(other, name))

    @classmethod
    def load_from_env(cls: _t.Type[_Self]) -> _Self:
        """Load config from environment variables.

        Use :meth:`Config.update` to merge several loaded configs into one.

        """

        fields = {}

        for name, field in cls._fields.items():
            if field.env is _DISABLED:
                continue

            if field.env in os.environ:
                fields[name] = field.parser(os.environ[field.env])

        return cls(**fields)

    @classmethod
    def load_from_args(
        cls: _t.Type[_Self],
        args: _t.Optional[argparse.Namespace] = None
    ) -> _Self:
        """Load config from parsed command line arguments.

        This method assumes that arguments parser was configured
        with :meth:`Config.setup_parser`.

        Use :meth:`Config.update` to merge several loaded configs into one.

        If `args` is not given, will create a parser and parse arguments
        from :data:`sys.argv`.

        """

        if args is None:
            args = cls.setup_parser().parse_args()

        fields = {}

        for name, field in cls._fields.items():
            if field.flag is _DISABLED:
                continue

            value = getattr(args, field.dest, _MISSING)
            if value is not _MISSING:
                fields[name] = value

        return cls(**fields)

    @classmethod
    def load_from_config(
        cls: _t.Type[_Self],
        config: _t.Dict[str, _t.Any],
        ignore_unknown_fields: bool = False
    ) -> _Self:
        """Load config from parsed config file.

        This method takes a dict with arbitrary values that resulted from
        parsing type-rich configs such as ``.yaml`` or ``.json``.

        For example::

            with open('conf.yml') as file:
                config = Config.load_from_config(yaml.load(file))

        Use :meth:`Config.update` to merge several loaded configs into one.

        """

        fields = {}

        if not ignore_unknown_fields:
            for name in config:
                if name not in cls._fields:
                    raise ValueError(f'unknown config field {name}')

        for name, field in cls._fields.items():
            value = config.get(name, _MISSING)
            if value is not _MISSING:
                value = field.parser.parse_config(value)
                field.parser.validate(value)
                fields[name] = value

        return cls(**fields)

    @classmethod
    def load_from_json(
        cls,
        path: str,
        ignore_unknown_fields: bool = False
    ):
        """Load config from a ``.json`` file.

        """

        with open(path, 'r') as file:
            try:
                loaded = json.load(file)
            except json.JSONDecodeError as e:
                raise ValueError(f'invalid config {path}: {e}')

        return cls.load_from_config(loaded, ignore_unknown_fields)

    @classmethod
    def setup_parser(
        cls,
        parser: _t.Optional[argparse.ArgumentParser] = None
    ) -> argparse.ArgumentParser:
        """Add fields from this config as flags to an argparse parser.

        If parser is not given, will create one.

        """

        if parser is None:
            parser = argparse.ArgumentParser()

        for name, field in cls._fields.items():
            if field.flag is _DISABLED:
                continue

            metavar = field.parser.describe()
            if metavar:
                metavar = '{' + metavar + '}'
            else:
                metavar = '<' + name.replace('_', '-') + '>'

            parser.add_argument(
                *field.flag,
                type=field.parser,
                default=_MISSING,
                help=field.help,
                metavar=metavar,
                dest=field.dest
            )

        return parser

    def __getattribute__(self, item):
        value = super().__getattribute__(item)
        if value is _MISSING:
            raise AttributeError(f'{item} is not configured')
        else:
            return value

    def __repr__(self):
        field_reprs = []
        for name in self._fields:
            value = getattr(self, name, _MISSING)
            if value is not _MISSING:
                field_reprs.append(f'{name}={value!r}')
        if field_reprs:
            field_desc = ",\n  ".join(field_reprs)
            return f'{self.__class__.__name__}(\n  {field_desc}\n)'
